from pathlib import Path
from collections import OrderedDict
import asyncio
import functools

import numpy as np
from lammps.output import LammpsRun, LammpsData
from lammps.inputs import LammpsInput, LammpsScript
from lammps.calculator.client import LammpsLocalClient

from .base import MDReader, MDCalculator
from .utils import element_type_to_symbol


class LammpsReader(MDReader):
    def __init__(self, directory, data_filename='initial.data', log_filename='lammps.log', dump_filename='mol.lammpstrj'):
        self.directory = Path(directory)
        self.data_filename = data_filename
        self.log_filename = log_filename
        self.dump_filename = dump_filename
        self._parse()

    def _parse(self):
        dump_path = list(self.directory.glob('**/%s' % self.dump_filename))
        if len(dump_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.dump_filename)
        elif len(dump_path) == 0:
            raise ValueError('could not find %s file within directory' % self.dump_filename)

        log_path = list(self.directory.glob('**/%s' % self.log_filename))
        if len(log_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.log_filename)
        elif len(log_path) == 0:
            raise ValueError('could not find %s file within directory' % self.log_filename)

        data_path = list(self.directory.glob('**/%s' % self.data_filename))
        if len(data_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.data_filename)
        elif len(data_path) == 0:
            raise ValueError('could not find %s file within directory' % self.data_filename)

        data = LammpsData.from_file(str(data_path[0]))
        output = LammpsRun(str(data_path[0]), lammps_log=str(log_path[0]), lammps_dump=str(dump_path[0]))
        # TODO: check units
        self._forces = output.get_forces(-1)
        self._stress = output.get_stress(-1)
        self._energy = output.get_energy(-1)
        self._structure = output.get_structure(-1)


lammps_dftfit_set = OrderedDict([
    ('echo', 'both'),
    ('log', 'lammps.log'),
    ('units', 'metal'),
    ('dimension', 3),
    ('boundary', 'p p p'),
    ('atom_style', 'full'),
    ('read_data', 'initial.data'),
    ('kspace_style', []),
    ('pair_style', []),
    ('pair_coeff', []),
    ('set', []),
    ('dump', '1 all custom 1 mol.lammpstrj id type x y z fx fy fz'),
    ('dump_modify', '1 sort id'),
    ('thermo_modify', 'flush yes'),
    ('thermo_style', 'custom step etotal pxx pyy pzz pxy pxz pyz'),
    ('run', 0),
])


def modify_input_for_potential(lammps_input, potential):
    symbol_indicies = {element_type_to_symbol(s): i for s, i in lammps_input.lammps_data.symbol_indicies.items()}

    def charge(potential):
        spec = potential.schema['spec']
        set_commands = []
        for element, charge in spec.get('charge', {}).items():
            set_commands.append('type %d charge %f' % (symbol_indicies[element], float(charge)))
        return ('set', set_commands)

    def kspace_style(potential):
        spec = potential.schema['spec']
        if 'kspace' in spec:
            style = spec['kspace']['type']
            tollerance = spec['kspace']['tollerance']
            return ('kspace_style', '%s %f' % (style, float(tollerance)))
        return ('kspace_style', [])

    def tersoff_file(potential):
        spec = potential.schema['spec']
        if 'pair' in spec and spec['pair']['type'] == 'tersoff':
            lines = []
            for parameter in spec['pair']['parameters']:
                lines.append(' '.join(parameter['elements'] + [float(c) for c in parameter['coefficients']]))
            return ('\n'.join(lines), 'potential.tersoff')
        return []

    def pair_style(potential):
        pair_map = {
            'buckingham': ('buck', '{style} {cutoff}'),
            'tersoff': ('tersoff', '{style}')
        }

        spec = potential.schema['spec']
        if 'pair' in spec:
            style, pair_style_format = pair_map[spec['pair']['type']]
            cutoff = spec['pair'].get('cutoff', 10) # angstroms
            if 'kspace'in spec and spec['kspace']['type'] in {'ewald', 'pppm'}:
                style += '/coul/long'
            return ('pair_style', pair_style_format.format(**{'style': style, 'cutoff': float(cutoff)}))
        return ('pair_style', [])

    def pair_coeff(potential):
        spec = potential.schema['spec']
        if 'pair' in spec:
            if spec['pair']['type'] in ['buckingham']:
                symbols_to_indicies = lambda symbols: [symbol_indicies[s] for s in symbols]
                pair_coeffs = []
                for coeff in spec['pair']['parameters']:
                    pair_coeffs.append(' '.join(list(map(str, symbols_to_indicies(coeff['elements']) + coeff['coefficients']))))
                return ('pair_coeff', pair_coeffs)
            elif spec['pair']['type'] == 'tersoff':
                return '* * potential.tersoff Si C Si'; # TODO: find out how to determine
        return ('pair_coeff', [])

    lammps_input.lammps_script.update([
        charge(potential),
        kspace_style(potential),
        pair_style(potential),
        pair_coeff(potential)
    ])
    lammps_input.additional_files.extend(tersoff_file(potential))


class LammpsLocalCalculator(MDCalculator):
    def __init__(self, command='lammps_ubuntu', num_workers=1):
        self.num_cores = 1
        self.command = command
        self.lammps_local_client = LammpsLocalClient(command=command, num_workers=num_workers)

    async def create(self):
        await self.lammps_local_client.create()

    async def submit(self, structure, potential, properties=None, lammps_set=None):
        lammps_set = lammps_set or lammps_dftfit_set
        properties = properties or {'stress', 'energy', 'forces'}
        lammps_input = LammpsInput(
            LammpsScript(lammps_set),
            LammpsData.from_structure(structure))
        modify_input_for_potential(lammps_input, potential)
        data_filename = lammps_input.lammps_script.data_filenames[0]
        stdin = str(lammps_input.lammps_script)
        files = {data_filename: str(lammps_input.lammps_data)}
        return await self.lammps_local_client.submit(stdin, files, properties)

    def shutdown(self):
        self.lammps_local_client.shutdown()



from tempfile import TemporaryDirectory
import os

import numpy as np
from pymatgen import Lattice, Specie, Structure

from dftfit.io import LammpsReader


def test_lammps_reader():
    lammps = LammpsReader('test_files/lammps/mgo')
    assert np.all(np.isclose(lammps.forces, np.zeros((8, 3))))
    assert np.all(np.isclose(lammps.stress, np.eye(3) * 661.01685))
    assert np.isclose(lammps.energy, -85.34101)
    structure = lammps.structure
    assert len(structure) == 8
    assert set(s.symbol for s in structure.species) == {'Mg', 'O'}



# def test_lammps_writer():
#     supercell = (5, 5, 5)
#     a = 4.1990858 # From evaluation of potential
#     lattice = Lattice.from_parameters(a, a, a, 90, 90, 90)
#     mg = Specie('Mg', 1.4)
#     o = Specie('O', -1.4)
#     atoms = [mg, o]
#     sites = [[0, 0, 0], [0.5, 0.5, 0.5]]
#     structure = Structure.from_spacegroup(225, lattice, atoms, sites)

#     potential_schema = {
#         'version': 'v1',
#         'kind': 'Potential',
#         'spec': {
#             'charge': {
#                 'Mg': 1.4, 'O': -1.4
#             },
#             'kspace': {
#                 'type': 'pppm', 'tollerance': 1e-5
#             },
#             'pair': {
#                 'type': 'buckingham',
#                 'cutoff': 10.0,
#                 'parameters': [
#                     {
#                         'elements': ['Mg', 'Mg'],
#                         'coefficients': [1309362.2766468062, 0.104, 0.0]
#                     },
#                     {
#                         'elements': ['Mg', 'O'],
#                         'coefficients': [9892.357, 0.20199, 0.0]
#                     },
#                     {
#                         'elements': ['O', 'O'],
#                         'coefficients': [2145.7345, 0.3, 30.2222]
#                     }
#                 ]
#             }
#         }
#     }

#     lammps = LammpsWriter(structure, potential_schema)
#     with TemporaryDirectory() as tempdir:
#         lammps.write_input(tempdir)
#         assert set(os.listdir(tempdir)) == {'lammps.in', 'initial.data'}



import asyncio
import os
import json

import numpy as np
from lammps.inputs import LammpsScript
from lammps.sets import MODULE_DIR
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core import Lattice
from pymatgen.analysis.elasticity import DeformedStructureSet, ElasticTensor, Stress, Strain

from ..io.lammps import LammpsLocalCalculator


def load_lammps_set(config_filename):
    with open(os.path.join(MODULE_DIR, 'sets', config_filename + ".json")) as f:
        return json.load(f, object_pairs_hook=LammpsScript)


class Predict:
    def __init__(self, calculator='lammps', loop=None, **kwargs):
        calculator_mapper = {
            'lammps': LammpsLocalCalculator
        }
        self.calculator = calculator_mapper[calculator](**kwargs)
        self.loop = loop or asyncio.get_event_loop()
        self.loop.run_until_complete(self.calculator.create())

    def conventional_structure(self, structure):
        sga = SpacegroupAnalyzer(structure)
        return sga.get_conventional_standard_structure()

    def static(self, structure, potential):
        async def calculate():
            future = await self.calculator.submit(
                structure, potential,
                properties={'forces', 'stress', 'energy'},
                lammps_set=load_lammps_set('static'))
            await future
            return future.result()
        result = self.loop.run_until_complete(calculate())
        return {
            'energy': result['results']['energy'],
            'stress': np.array(result['results']['stress']) * 1e-4,  # convert to GPa
            'forces': np.array(result['results']['forces'])
        }

    def lattice_constant(self, structure, potential, supercell=(1, 1, 1), etol=1e-6, ftol=1e-6):
        conventional_structure = self.conventional_structure(structure)

        async def calculate():
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['minimize'] = '%f %f 2000 10000' % (etol, ftol)
            future = await self.calculator.submit(
                conventional_structure * supercell, potential,
                properties={'lattice'},
                lammps_set=relax_lammps_script)
            await future
            return future.result()

        result = self.loop.run_until_complete(calculate())
        lattice = Lattice(result['results']['lattice'])
        return conventional_structure.lattice, Lattice(lattice.matrix / np.array(supercell))

    def elastic_constant(self, structure, potential, supercell=(1, 1, 1),
                         nd=0.01, ns=0.05, num_norm=4, num_shear=4,
                         etol=1e-6, ftol=1e-6):
        conventional_structure = self.conventional_structure(structure)
        deformation_set = DeformedStructureSet(conventional_structure * supercell,
                                               nd=nd, ns=ns,
                                               num_norm=num_norm, num_shear=num_shear)
        strains = []
        stresses = []

        async def calculate():
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['fix'] = []
            relax_lammps_script['minimize'] = '%f %f 2000 10000' % (etol, ftol)
            futures = []
            for deformation, deformed_structure in zip(deformation_set.deformations, deformation_set):
                strains.append(Strain.from_deformation(deformation))
                futures.append(await self.calculator.submit(
                    deformed_structure, potential,
                    properties={'stress'},
                    lammps_set=relax_lammps_script))
            for result in await asyncio.gather(*futures):
                stress = Stress(np.array(result['results']['stress']) * 1e-4) # Convert to GPa
                stresses.append(stress)

        self.loop.run_until_complete(calculate())
        return ElasticTensor.from_pseudoinverse(strains, stresses)
